"""
analytics classes and functions related to the simulation

All functions are included on the Analytics object, so you can import all of them as

    from sim_analytics import Analytics as A
    A.linspace(100)

Most functions are also exposed at module level, so alternatively you can do

    import sim_analytics as al
    al.linspace(100)


(c) Copyright Bprotocol foundation 2022. 
Licensed under MIT
"""
__version__ = "1.2"
__date__ = "24/Nov/2022"

import numpy as np
import pandas as pd
from collections import namedtuple
from ..pair import CarbonPair
from matplotlib import pyplot as plt
from math import isnan
    

class Analytics:
    """
    collection of analytics functions useful for Carbon simulations

    :sim:       the CarbonSimulatorUI object

    NOTE: this library relies heavily on numpy, and all vectors are in fact numpy arrays
    """
    def __init__(self, sim=None, verbose=False):
        self.sim = sim
        self.verbose = verbose

    @staticmethod
    def midpoints(r):
        """
        calculates the midpoints: (x0, x1, x2, ...) -> (avg(x0,x1), avg(x1,x2), ...)
        """
        return Analytics.vec([0.5*(x1+x2) for x1,x2 in zip(r, r[1:])])

    @staticmethod
    def diff(vec):
        """
        calculates the differences: (x0, x1, x2, ...) -> (x1-x0, x2-x1, ...)
        """
        return Analytics.vec(np.diff(vec))
        # return np.diff(vec)

    @staticmethod
    def vec(v):
        """
        converts iterable into vector*

        *currently our vector format is np.array; however, this is NOT guaranteed and we may
        switch over to a difference vector implementation the allows for basic vector operations
        (component-wise +-*/ using the usual syntax)
        """
        return pd.Series(v)
        # return np.array(v)

    @staticmethod
    def vecdot(v1, v2):
        """implements the dot product (sumproduct)"""
        return v1.fillna(0).dot(v2.fillna(0))
        # try:
        #     return np.dot(v1, v2)
        # except TypeError:
        #     v1a = [0 if x is None else x for x in v1]
        #     v2a = [0 if x is None else x for x in v2]
        #     print(f"[vecdot] typeerror {v1} {v2}; using {v1a} {v2a}")
        #     return np.dot(v1a, v2a)
            
    EPSILON = 0.0001
    EPSILON2 = 0.00001
    @classmethod
    def linspace(cls, xmax, nint=10, eps=None, eps2=None):
        """
        generates a linearly spaced vector of floats (0+eps, x1, x2, ... xmax*(1-eps2))

        :xmax:          maximum x to be generated
        :nint:          number of intervals, ie numper of points - 1

        The function generates nint equally space intervals (nint+1 points), starting
        from ~0 and ending at ~xmax. The intermediate points are based on the exact
        interval (0, xmax). Once the series has been generated to following adjustments
        are made for numerical stability reasons

            vec[0]  = eps
            vec[-1] = (1-eps2) * maxx
        """
        if eps is None: eps = cls.EPSILON
        if eps2 is None: eps2 = cls.EPSILON2
        vec = np.linspace(0,xmax, nint+1)
        vec[0] = eps
        vec[-1] = (1-eps2) * xmax
        return cls.vec(vec)
        # return vec

    @staticmethod
    def isnonenan(x):
        """
        returns true if None or Nan, False else
        """
        if x is None: return True
        try:
            return isnan(x)
        except:
            return False
    
    @classmethod
    def truncate(cls, s1, s2):
        """
        truncates both series to the same length, removing NaN / None from the end of the second
        
        :s1:                shortened if need be to the length of s2
        :s2:                truncated of NaN / None at the end
        :returns:           truncated (s1, s2)
        """
        s2r = cls.vec(x for x in s2 if not cls.isnonenan(x))
        s1r = s1[:len(s2r)]
        return s1r, s2r

    BID_BOOK = "amm_buys"
    BID = BID_BOOK
    ASK_BOOK = "amm_sells"
    ASK = ASK_BOOK
    OFFER_BOOK = ASK_BOOK
    OFFER = ASK_BOOK
    MAXPRICE = 1E100        # price returned if out of liquidity for amm_buys
    
    def effective_price(self, size, bidask, tkn=None, sim=None, pair=None):
        """
        returns the price at which `tkn` can be bought or sold in `size`

        :size:          the amount of tkn being bought or sold
        :bidask:        whether  to look at the bid book (AMM buys) or the ask book (AMM sells)
                        use the constants BID=BID_BOOK, ASK=ASK_BOOK
        :tkn:           the token being bought or sold (usually the base token of the pair)
                        can be omitted if the sim is provided and has been seeded with CarbonPair
        :sim:           the CarbonSimulatorUI object (from self.sim if None)
        :pair:          the corresponding pair (if none, use sim defaults)
        """
        if sim is None: sim = self.sim

        if tkn is None:
            if sim is None:
                raise ValueError("If tkn is None then sim must be provided in arguments or class defaults")
            tkn = sim.default_basetoken
            if tkn is None:
                raise ValueError(f"The sim provided does not specify a basetoken [{sim}]")

        if pair is None: 
            if sim is None:
                raise ValueError("If pair is None then sim must be provided in arguments or class defaults")
            pair = sim.pair 
        if bidask == self.BID_BOOK: 
            price_f = sim.amm_buys
            fail_value = self.MAXPRICE
        elif bidask == self.ASK_BOOK: 
            price_f = sim.amm_sells
            fail_value = 0.
        else:
            raise ValueError(f"Illegal value for bidask [{bidask}]")
        
        try:
            result = price_f(tkn, size, execute=False, pair=pair)
            price = result["trades"].iloc[-1]["price"]
            if self.verbose:
                print(f"[effective_price] trading size={size} price={price} [pair={pair}, tkn={tkn} {bidask}]")
            return float(price)
        except:
            return fail_value


    AMM_SELLS_BASETOKEN = "amm_sells_tknb"
    ASK = "amm_sells_tknb"
    AMM_BUYS_BASETOKEN = "amm_buys_tknb"
    BID = "amm_buys_tknb"
    
    def simulate_trades(self, src_size, buysell, sim=None, carbon_pair=None):
        """
        returns amount against which `tkn` can be traded (bought or sold) or sold in `src_size`

        :src_size:          the amount of the source tkn being sold or bought (always positive!)
        :buysell:           either AMM_SELLS_BASETOKEN or AMM_BUYS_BASETOKEN
        :sim:               the CarbonSimulatorUI object (from self.sim if None)
        :carbon_pair:       the corresponding pair as CarbonPair object (if none, use sim defaults)
        """
        if sim is None: sim = self.sim
        if carbon_pair is None:
            if self.sim is None:
                raise ValueError("The `sim` object must be provided and have a default carbon pair if `carbon_pair` is not given.")
            sim = self.sim
            if sim.carbon_pair is None:
                raise ValueError("The `sim` object must have a default carbon pair if `carbon_pair` is not given.")
            carbon_pair = sim.carbon_pair
        if not isinstance(carbon_pair, CarbonPair):
            raise ValueError("The `carbon_pair` provided is not a CarbonPair instance", carbon_pair)

        pair = carbon_pair.pair_iso
        tkn = carbon_pair.tknb
        if buysell == self.AMM_SELLS_BASETOKEN:
            func = sim.amm_sells
            fail_value = None
            ix = "amt2"
        elif buysell == self.AMM_BUYS_BASETOKEN:
            func = sim.amm_buys
            fail_value = None
            ix = "amt1"
        else:
            raise ValueError(f"Unknonwn buysell {buysell}")
        try:
            result = func(tkn, src_size, execute=False, pair=pair)
            trg_size = result["trades"].iloc[-1][ix]
            result = trg_size
        except:
            result = fail_value
        if self.verbose:
            print(f"[simulate_trades] trading src={src_size} trg={result} [pair={pair}, tknb={tkn} {buysell}]")
        return result
            
    def __repr__(self):
        return f"{self.__class__.__name__}(sim={self.sim})"

orders_nt = namedtuple("orders_nt", 'tkn, amt, p_start, p_end')
midpoints = Analytics.midpoints
linspace = Analytics.linspace
diff = Analytics.diff
vec = Analytics.vec
vecdot = Analytics.vecdot
truncate = Analytics.truncate
isnonenan = Analytics.isnonenan

class OrderBook():
    """
    this class represents a (numerical) order book
    
    :src_amounts:       a vector of trade sizes (`src_tkn`; always positive)
    :trg_amounts:       a vector of tokens received in those trades (`trg_tkn`; alwaus positive)
    :src_tkn:           the name of the source token (ie the one appearing on the x-axis driving the deal)
    :trg_tkn:           ditto target token (ie the other one, the one that is reacting)



    The OrderBook class in particular prepares the data to draw a number
    of charts, eg with the matplotlib command

        plt.plot(x, y)

    Below we indicate which properties provide the data for which chart
    (we assume the OrderBook object is called `ob`)

    - Token Amount chart: plots the amount of token in vs out
        :x:     ob.src_amounts
        :y:     ob.trg_amounts

    - Price chart: plots the effective (1) and marginal (2) prices vs out amount
        :x1:    ob.src_amounts
        :y1:    ob.prices
        :x2:    ob.midpoints(ob.src_amounts)
        :y2:    ob.marg_prices

    - Order Book chart: plots the order book (liquidity vs price)
        :x:     ob.ob_liquidity
        :y:     ob.ob_prices

    """
    ASK = "ASK_amm_sells_src_tkn"
    BID = "BID_amm_buys_src_tkn"
    def __init__(self, src_amounts, trg_amounts, src_tkn="SRC", trg_tkn="TRG", bidask=ASK):
        src_amounts, trg_amounts = Analytics.truncate(src_amounts, trg_amounts)
        self.src_amounts = Analytics.vec(src_amounts)
        self.trg_amounts = Analytics.vec(trg_amounts)
        self.src_tkn = src_tkn
        self.trg_tkn = trg_tkn
        self.bidask = bidask
        if bidask == self.ASK:
            self.reverse_book = False
        elif bidask == self.BID:
            self.reverse_book = True
        else:
            raise ValueError(f"illegal value for bidask {bidask}")

    @property
    def base_tkn(self):
        """returns the base token (=self.src_tkn)"""
        return self.src_tkn

    @property
    def quote_tkn(self):
        """returns the quote token (=self.trg_tkn)"""
        return self.trg_tkn

    def explain(self):
        """
        human-readable explanation 
        """
        ba    = 'ASK' if self.bidask==self.ASK else 'BID'
        msg0  = f"This is the {ba} book."
        msg1  = f"Source token = {self.src_tkn}, target token = {self.trg_tkn}."
        bs    = "sells" if self.bidask == self.ASK else "buys"
        msg2  = f"AMM {bs} {self.src_tkn} for {self.trg_tkn}."
        msg3  = f"Base token = {self.base_tkn}, quote token = {self.quote_tkn}."
        msg3b = f"Prices are quoted in {self.prices_u}."
        msg4  = f"Order book amounts are quoted in {self.ob_liquidity_u}."
        return "\n".join([msg0, msg1, msg2, msg3, msg3b, msg4])

    @property
    def src_token_outin(self):
        """'out' if source token goes out, 'in' otherwise"""
        return "in" if self.reverse_book else "out"

    @property
    def trg_token_outin(self):
        """'out' if target token goes out, 'in' otherwise"""
        return "out" if self.reverse_book else "in"

    @property
    def amm_buysell_src(self):
        """'sells' if AMM sells source token, 'buys' otherwise"""
        return "buys" if self.reverse_book else "sells"

    @property
    def amm_buysell_trg(self):
        """'sells' if AMM sells target token, 'buys' otherwise"""
        return "sells" if self.reverse_book else "buys"

    @property
    def amm_bidask(self):
        """'ask' if AMM sells target token, 'bid' otherwise"""
        return "bid" if self.reverse_book else "ask"

    @property
    def marg_trg_amounts(self):
        """
        the marginal in amounts (amt1-amt0, amt2-amt1, ...)
        """
        return Analytics.diff(self.trg_amounts)

    @property
    def marg_src_amounts(self):
        """
        the marginal out amounts (amt1-amt0, amt2-amt1, ...)
        """
        return Analytics.diff(self.src_amounts)

    @property
    def prices(self):
        """
        returns the effective prices (trg_amounts / src_amounts)
        """
        return self.trg_amounts / self.src_amounts
    
    def pricesr(self, reverse=True):
        """
        returns the reverse effective prices (1/prices) if reverse True (default), otherwise prices
        """
        return 1/self.prices if reverse else self.prices
    
    @property
    def marg_prices(self):
        """
        returns the reverse marginal prices (marg_src_amounts / marg_trg_amounts)
        """
        return self.marg_trg_amounts / self.marg_src_amounts 
    
    def marg_pricesr(self, reverse=True):
        """
        returns the reverse effective prices (1/marg_prices) if reverse True (default), otherwise marg_prices
        """
        return 1/self.marg_prices if reverse else self.marg_prices
    
    @property
    def prices_u(self):
        """
        unit of `prices`, `marg_prices`, etc
        """
        return f"{self.quote_tkn} per {self.base_tkn}"
    marg_prices_u = prices_u

    def pricesr_u(self, reverse=True):
        """
        unit of `pricesr`, `marg_pricesr`, etc
        """
        return f"{self.src_tkn} per {self.trg_tkn}" if reverse else self.prices_u
    marg_pricesr_u = pricesr_u
    
    @property
    def ob_liquidity(self):
        """
        returns the order book liquidity at the price points corresponding to `ob_prices`
        """
        return self.midpoints(self.marg_trg_amounts) / self.dmarg_prices

    @property
    def dmarg_prices(self):
        """
        returns the differences of marginal prices (dp1-dp0, dp2-dp1, ...)

        Note: the dot product `Analytics.vecdot(dmarg_prices, ob_liquidity)`, which corresponds
        to the area under the curve, is approximately the entire liquidity in the curve, 
        ie the same as `max(trg_amounts)`
        """
        result = Analytics.diff(self.marg_prices)
        if self.reverse_book:
            result = - result
        return result
    
    @property
    def ob_liquidity_u(self):
        """
        returns the units of ob_liquidity
        """
        return f"{self.trg_tkn}"

    @property
    def ob_prices(self):
        """
        returns the prices corresponding to the order book liquidity in `ob_liquidity`
        """
        return self.midpoints(self.marg_prices)
    ob_prices_u = prices_u

    @staticmethod
    def reverse(vec):
        """
        returns inverse of vec = (x0, x1, ..) -> (1/x0, 1/x1, ...)
        """
        return 1/vec
    inverse = reverse

    @staticmethod
    def midpoints(vec):
        """
        alias for Analytics.midpoints
        """
        return Analytics.midpoints(vec)
    mp = midpoints

    @staticmethod
    def calc_liquidity_approx(orderuis, prices, pair, reverse=True, purgezero=True):
        """
        calculates the approximate liquidity of all positions
        
        :orderuis:      the dict returned by Sim.state()["orderuis"]
                        
        :prices:        the price range over which to calculate the liquidity
                        eg np.linspace(500,3000, 100)
        :pair:          the pair (CarbonPair object) for which to calculate the liquidity
                        eg CarbonPair(tknb="ETH", tknq="USDC"); note that the curve 
                        price convention MUST be the same as the one in orderuis
        :reverse:       if True (default), calculate the liquidity in terms of quote token
                        (eg USDC); otherwise calculate it in terms of base token (eg ETH)
        :purgezero:     if True (default), replace zeroes with None
        :returns:       example {"ask":list, "bid":list, "prices":list, "pair":CarbonPair
                        "price_u":"USDC per ETH", "liqtkn_u":"USDC", "reverse": True}
        """
        if not isinstance(pair, CarbonPair):
            raise ValueError("Pair must be a CarbonPair", pair)
        
        pair_iso = pair.pair_iso
        tkn_ask = pair.tknb
        tkn_bid = pair.tknq
        print(f"[calc_liquidity_approx] pair:{pair_iso}", tkn_ask, tkn_bid)
        
        curves_ask = [
            r for r in orderuis.values() 
            if r.pair.pair_iso == pair_iso and r.tkn == tkn_ask
        ]
        curves_bid = [
            r for r in orderuis.values() 
            if r.pair.pair_iso == pair_iso and r.tkn == tkn_bid
        ]
        print(f"[calc_liquidity_approx] ask:{len(curves_ask)} bid:{len(curves_bid)}")
        
        tkn = pair.tknq if reverse else pair.tknb
        print(f"[calc_liquidity_approx] tkn={tkn}")
        liq_ask = [
            sum(r.liquidity_approx(p1, p2, tkn)/(p2-p1) for r in  curves_ask) 
            for p1, p2 in zip(prices, prices[1:])
        ]
        liq_bid = [
            sum(r.liquidity_approx(p1, p2, tkn)/(p2-p1) for r in  curves_bid) 
            for p1, p2 in zip(prices, prices[1:])
        ]
        pp = [0.5*(p1+p2) for p1, p2 in zip(prices, prices[1:])]
        
        if purgezero:
            liq_ask = [x if x else None for x in liq_ask]
            liq_bid = [x if x else None for x in liq_bid]
            
        return {
            "bid":      liq_bid, 
            "ask":      liq_ask,
            "prices":   pp,
            "pair":     pair,
            "price_u":  pair.price_convention,
            "liqtkn_u": tkn,
            "reverse":  reverse,
        }



    COLOR_BID = "g"
    COLOR_ASK = "r"

    @property
    def colorba(self):
        """returns the correct color depending on bid/ask"""
        return self.COLOR_BID if self.bidask==self.BID else self.COLOR_ASK
    
    def plot_token_amount_chart(self):
        """
        plots the token amount chart based on the order book in self
        """
        ob=self
        x = ob.src_amounts
        y = ob.trg_amounts
        plt.plot(x, y, marker="", label="token amount")
        plt.title(f"Token trade (AMM {ob.amm_buysell_src} {ob.src_tkn}, {ob.amm_buysell_trg} {ob.trg_tkn})")
        plt.xlabel(f"Tokens {ob.src_token_outin} ({ob.src_tkn})")
        plt.ylabel(f"Tokens {ob.trg_token_outin} ({ob.trg_tkn})")
        plt.grid()
        #plt.legend()
        return f"plotted tokens received against trade size ({int(max(y)):,})"


    def plot_price_charto(self):
        """
        plots the price chart based on the order book in self
        """
        ob=self
        x1 = ob.src_amounts
        y1 = ob.prices
        plt.plot(x1, y1, marker="", label="effective price")
        x2 = ob.midpoints(x1)
        y2 = ob.marg_prices
        plt.plot(x2, y2, marker="", label="marginal price")
        plt.title("Price against trade size")
        plt.xlabel(f"Trade size ({ob.src_tkn}; AMM {ob.amm_buysell_src})")
        plt.ylabel(f"Price ({ob.trg_tkn} per {ob.src_tkn})")
        plt.grid()
        plt.legend()
        return "plotted marginal and effective prices against trade size"

    def plot_price_chart(self, reverse=False):
        """
        plots the price chart based on the order book in self
        """
        ob=self
        x1 = ob.src_amounts
        y1 = ob.pricesr(reverse)
        plt.plot(x1, y1, marker="", label="effective price")
        x2 = ob.midpoints(x1)
        y2 = ob.marg_pricesr(reverse)
        plt.plot(x2, y2, marker="", label="marginal price")
        plt.title("Price against trade size")
        plt.xlabel(f"Trade size ({ob.src_tkn}; AMM {ob.amm_buysell_src})")
        plt.ylabel(f"Price ({ob.pricesr_u(reverse)})")
        plt.grid()
        plt.legend()
        return "plotted marginal and effective prices against trade size"

    def plot_orderbook_chart(self, xmin=None, xmax=None, ymax=None, otherob=None):
        """
        plots order book chart based on the order book in self
        
        :otherob:   if given, also plot the other orderbook into the same chart
        """
        ob=self
        y = ob.ob_liquidity
        x = ob.ob_prices
        plt.plot(x, y, marker=".", color=self.colorba, label=f"{ob.amm_bidask}")
        if otherob:
            yo = otherob.ob_liquidity
            xo = otherob.ob_prices
            plt.plot(xo, yo, marker=".", color=otherob.colorba, label=f"{otherob.amm_bidask}")

        fragment = f" ({ob.amm_bidask})" if not otherob else " (bid and ask)"
        plt.title(f"Order book{fragment}")
        plt.xlabel(f"Price ({ob.ob_prices_u})")
        plt.ylabel(f"Liquidity ({ob.ob_liquidity_u})")
        plt.grid()
        if otherob:
            plt.legend()
        plt.xlim(xmin,xmax)
        plt.ylim(0,ymax)
        return f"plotted order book ({int(Analytics.vecdot(ob.dmarg_prices,y)):,})"

    @staticmethod
    def plot_approx_orderbook_chart(l):
        """
        plots approximate order book chart
        
        :l:   result returned from calc_liquidity_approx()
        """
        plt.title(f"Approximate Order Book (tknb={l['pair'].tknb}, tknq={l['pair'].tknq})")
        plt.xlabel(f"Price ({l['price_u']})")
        plt.ylabel(f"Liquidity ({l['liqtkn_u']})")
        plt.plot(l['prices'], l['bid'], color="g", label="bid")
        plt.plot(l['prices'], l['ask'], color="r", label="ask")
        plt.grid()
        plt.legend()

    def __repr__(self):
        return f"{self.__class__.__name__}(<src={self.src_tkn} trg={self.trg_tkn} bidask={self.bidask} rev={self.reverse_book}; {len(self.src_amounts)-1} records>)"

calc_liquidity_approx       = OrderBook.calc_liquidity_approx
plot_approx_orderbook_chart = OrderBook.plot_approx_orderbook_chart





"""
The Carbon Simulator, a simulation engine for the Carbon protcol

see https://carbondefi.xyz for details and more documentation
see https://github.com/bancorprotocol/carbon-simulator for the repo

(c) Copyright Bprotocol foundation 2022.
Licensed under MIT
"""
from .routers import (
    ExactRouterX0Y0N,
)
from .pair import CarbonPair
from .simulators import CarbonSimulatorUI
from .simulators import sim_analytics as analytics


__version__ = "1.3"
__date__ = "30/Nov/2022"



from carbon import CarbonSimulatorUI
import pandas as pd


def test_liquidity_as_dataframe():
    """
    Derived from `passed_tests` notebook test-34
    """
    Sim = CarbonSimulatorUI(verbose=True)
    Sim.add_order("ETH", 10, 2000, 3000, "ETHUSDC")
    Sim.add_order("ETH", 20, 2010, 3010, "ETHUSDC")
    Sim.add_order("ETH", 30, 2020, 3020, "ETHUSDC")
    Sim.add_order("ETH", 40, 2030, 2030, "ETHUSDC")
    Sim.add_strategy("ETH", 10, 2000, 3000, 5000, 1000, 900, "ETHUSDC")
    Sim.add_strategy("ETH", 20, 2010, 3010, 5100, 1010, 910, "ETHUSDC")
    Sim.add_strategy("ETH", 30, 2020, 3020, 5200, 1020, 920, "ETHUSDC")
    Sim.add_strategy("ETH", 40, 2030, 3030, 5300, 1030, 930, "ETHUSDC")
    Sim.add_order("ETH", 10, 2000, 3000, "ETHDAI")
    Sim.add_order("ETH", 20, 2010, 3010, "ETHDAI")
    Sim.add_order("ETH", 30, 2020, 3020, "ETHDAI")
    Sim.add_order("ETH", 40, 2030, 2030, "ETHDAI")
    Sim.add_strategy("ETH", 10, 2000, 3000, 5000, 1000, 900, "ETHDAI")
    Sim.add_strategy("ETH", 20, 2010, 3010, 5100, 1010, 910, "ETHDAI")
    Sim.add_strategy("ETH", 30, 2020, 3020, 5200, 1020, 920, "ETHDAI")
    Sim.add_strategy("ETH", 40, 2030, 3030, 5300, 1030, 930, "ETHDAI")
    Sim.add_sgl_pos("ETH", 10, 3000, 2000, pair="ETHUSDC")

    df = Sim.liquidity(Sim.ASDF)
    dct = df.to_dict(orient="dict")["y"]
    assert isinstance(df, pd.DataFrame)
    assert dct == {
        ("ETHDAI", "DAI"): 20600.0,
        ("ETHDAI", "ETH"): 200.0,
        ("ETHUSDC", "ETH"): 210.0,
        ("ETHUSDC", "USDC"): 20600.0,
    }

    pairs = set(k[0] for k in dct)
    assert pairs == {"ETHDAI", "ETHUSDC"}
    assert {p: {k[1]: 1 for k in dct if k[0] == p} for p in pairs} == {
        "ETHUSDC": {"ETH": 1, "USDC": 1},
        "ETHDAI": {"DAI": 1, "ETH": 1},
    }


def test_liquidity_as_dict():
    """
    Derived from `passed_tests` notebook test-34
    """
    Sim = CarbonSimulatorUI(verbose=True)
    Sim.add_order("ETH", 10, 2000, 3000, "ETHUSDC")
    Sim.add_order("ETH", 20, 2010, 3010, "ETHUSDC")
    Sim.add_order("ETH", 30, 2020, 3020, "ETHUSDC")
    Sim.add_order("ETH", 40, 2030, 2030, "ETHUSDC")
    Sim.add_strategy("ETH", 10, 2000, 3000, 5000, 1000, 900, "ETHUSDC")
    Sim.add_strategy("ETH", 20, 2010, 3010, 5100, 1010, 910, "ETHUSDC")
    Sim.add_strategy("ETH", 30, 2020, 3020, 5200, 1020, 920, "ETHUSDC")
    Sim.add_strategy("ETH", 40, 2030, 3030, 5300, 1030, 930, "ETHUSDC")
    Sim.add_order("ETH", 10, 2000, 3000, "ETHDAI")
    Sim.add_order("ETH", 20, 2010, 3010, "ETHDAI")
    Sim.add_order("ETH", 30, 2020, 3020, "ETHDAI")
    Sim.add_order("ETH", 40, 2030, 2030, "ETHDAI")
    Sim.add_strategy("ETH", 10, 2000, 3000, 5000, 1000, 900, "ETHDAI")
    Sim.add_strategy("ETH", 20, 2010, 3010, 5100, 1010, 910, "ETHDAI")
    Sim.add_strategy("ETH", 30, 2020, 3020, 5200, 1020, 920, "ETHDAI")
    Sim.add_strategy("ETH", 40, 2030, 3030, 5300, 1030, 930, "ETHDAI")
    Sim.add_sgl_pos("ETH", 10, 3000, 2000, pair="ETHUSDC")

    dic = Sim.liquidity(Sim.ASDICT)
    assert isinstance(dic, dict)
    assert dic == {
        "ETHDAI": {"DAI": 20600.0, "ETH": 200.0},
        "ETHUSDC": {"ETH": 210.0, "USDC": 20600.0},
    }



from carbon import CarbonSimulatorUI


def test_unrelated_token_in_add_order():
    Sim = CarbonSimulatorUI(pair="ETHUSDC", verbose=False, raiseonerror=False)

    response = Sim.add_order("LINK", 100, 2000, 2000)
    assert response["success"] is False
    assert (
        response["error"]
        == "('Invalid token specification (tkn not part of isopair)', 'ETHUSDC', 'LINK')"
    )

    response = Sim.add_order("DNE", 100, 2000, 2000)
    assert response["success"] is False
    assert (
        response["error"]
        == "('Invalid token specification (tkn not part of isopair)', 'ETHUSDC', 'DNE')"
    )



"""
Base class for all routers.

(c) Copyright Bprotocol foundation 2022.
Licensed under MIT
"""
import decimal
import json
import math
from decimal import Decimal
from abc import abstractmethod
from typing import List, Tuple, Callable
from dataclasses import field, dataclass

from ..order import Order, DecFloatInt
import logging


@dataclass
class Rate:
    """
    Represents a rate between two tokens for the FastRouter.
    """

    input: DecFloatInt = 0
    output: DecFloatInt = 0


@dataclass
class Quote:
    """
    Represents a quoted Rate in the FastRouter.
    """

    index: int = 0
    rate: Rate = Rate()


@dataclass
class Action:
    """
    Represents an action response from the FastRouter.
    """

    index: int = 0
    x: DecFloatInt = 0
    input: DecFloatInt = 0
    output: DecFloatInt = 0
    price: DecFloatInt = 0
    total_output: DecFloatInt = 0
    total_input: DecFloatInt = 0
    total_price: DecFloatInt = 0
    match_method: str = ""
    threshold_orders: int = 100


@dataclass
class BaseRouter:
    """
    Base class for both Simple and Complex routers.
    """

    # A list of all curves for a given token pair.
    orders: List[Order] = field(default_factory=list)

    # Allows for toggling between floor division vs regular division
    use_floor_division: bool = False

    # Verbosity mode for logging.
    verbose: bool = False

    # Extra verbose logging mode.
    debug: bool = False

    # if False (default), errors are caught and returned in the result dict
    raiseonerror: bool = False

    assert_precision: int = 4

    @property
    def indexes(self) -> list:
        """
        Alias for position_list (used for shorthand and docs consistency).
        """
        return list(range(len(self.orders)))

    @staticmethod
    def frmt(val: DecFloatInt, num_decimals=12) -> float:
        """
        Values based on linear modelling of Barak gas estimates 20221003
        """
        return round(float(val), num_decimals)

    @staticmethod
    def prod(lst: List[DecFloatInt]) -> DecFloatInt:
        """
        Multiplies all elements in a given list in such a way that supports returning a Decimal data type.
        """
        return math.prod(lst)

    def sufficient_liquidity_exists(self, x: DecFloatInt, use_positions_matchlevel) -> bool:
        """
        Method to check if there is sufficient liquidity to handle the trade.
        """
        if use_positions_matchlevel == []:
            use_positions_matchlevel = self.indexes
        available_liquidity = sum([self.orders[i].y for i in use_positions_matchlevel])
        try:
            assert (
                available_liquidity - abs(x) >= 0
            ), "Insufficient liquidity across all user positions to support this trade."
            return True
        except AssertionError as e:
            raise e

    @abstractmethod
    def amt_by_target(
        self, subject: int, dx: DecFloatInt, position_subset: List[int] = None
    ) -> DecFloatInt:
        """
        Enforced method to handle getting the input amount for a given target amount.
        """
        pass

    @abstractmethod
    def amt_by_src(
        self, subject: int, dx: DecFloatInt, position_subset: List[int] = None
    ) -> DecFloatInt:
        """
        Enforced method to handle getting the target amount for a given source amount.
        """
        pass

    @abstractmethod
    def match(
        self,
        x: DecFloatInt,
        is_by_target: bool = True,
        completed_trade: bool = False,
        trade: Callable = None,
        cmp: Callable = None,
        check_sufficient_liquidity: bool = True,
        threshold_orders: int = 100,
    ) -> List[Action]:
        """
        Main algorithm to handle matching a trade amount against the curves/orders.
        """
        pass

    @abstractmethod
    def match_by_src(
        self,
        x: DecFloatInt,
        is_by_target: bool = False,
        check_sufficient_liquidity: bool = True,
        threshold_orders: int = 100,
    ) -> List[Action]:
        """
        Alias for match method in the case of a source amount.
        """
        pass

    @abstractmethod
    def match_by_target(
        self,
        x: DecFloatInt,
        is_by_target: bool = True,
        check_sufficient_liquidity: bool = True,
        threshold_orders: int = 100,
    ) -> List[Action]:
        """
        Alias for match method in the case of a target amount.
        """
        pass

    @staticmethod
    def from_json(json_file_path: str) -> dict:
        """
        Method to load a json file.
        """
        with open(json_file_path, "r") as f:
            return json.load(f)

    @staticmethod
    def from_csv(csv_file_path: str) -> dict:
        """
        Method to load a json file.
        """
        import pandas as pd

        return pd.read_csv(csv_file_path)

    def make_orders_from_file(
        self,
        file_path: str,
        format_type: str = "nix",
        num_curves: int = None,
        file_type: str = "json",
    ) -> Tuple[List[Order], dict] or List[Order]:
        """
        Method to make curves from a json file.
        """
        if num_curves is None:
            num_curves = 10

        curves = []
        nix_format = {}
        ct = 0
        if file_type == "json":
            curve_data = BaseRouter.from_json(file_path)
            for curve in curve_data:
                if type(curve) is not dict:
                    curve = curve_data[curve]
                try:
                    curve = {
                        k: Decimal(v)
                        for k, v in curve.items()
                        if k in ["S", "B", "_y", "y_int"]
                    }
                except decimal.InvalidOperation as e:
                    self.logger.error("decimal.InvalidOperation: ", type(curve), curve)
                    raise e
                c = Order(**curve)
                if self.use_floor_division:
                    p_high = int(float(Decimal(c.p_high).sqrt())) * 2**32
                    p_low = int(float(Decimal(c.p_low).sqrt())) * 2**32
                    y_int = c.y_int
                    y = c.y
                    c = Order(
                        p_low=c.p_low,
                        p_high=c.p_high,
                        S=p_high - p_low,
                        B=p_low,
                        y_int=y_int,
                        _y=y,
                        auto_convert_variables=False,
                    )
                curves.append(c)
                nix_format[ct] = {
                    "token_1": "usdc",
                    "token_2": "eth",
                    "token_1_p_a": c.p_high,
                    "token_1_p_b": c.p_low,
                    "token_1_y_int": c.y_int,
                    "token_2_p_a": c.p_high,
                    "token_2_p_b": c.p_low,
                    "token_2_y_int": c.y_int,
                }
                ct += 1
                if len(curves) == num_curves and num_curves is not None:
                    break
            if num_curves > len(curves):
                self.logger.warning(
                    f"num_cruves={num_curves} is greater than the number of curves found in the json_file_path..."
                    "using all curves found in the json_file_path."
                )
        elif file_type == "csv":
            curve_data = BaseRouter.from_csv(file_path)

            def unpack_row(indx):
                row = curve_data.iloc[indx]
                return {
                    "p_low": Decimal(float(row["p_low"])),
                    "p_high": Decimal(float(row["p_high"])),
                    "y_int": Decimal(float(row["y_int"])),
                }

            curve_data["index"] = [i for i in range(len(curve_data))]
            for i in curve_data["index"].values:
                curve = unpack_row(i)
                c = Order(**curve)
                curves.append(c)
            nix_format = {}
        if format_type == "nix":
            return curves, nix_format
        else:
            return curves

    def __post_init__(self):
        self.logger = logging
        if self.verbose:
            self.logger.basicConfig(level=logging.INFO)
        else:
            self.logger.basicConfig(level=logging.WARNING)



from carbon import CarbonSimulatorUI


def test_inverted_range_positions_1():
    """
    Derived from `passed_tests` notebook test-13
    """
    Sim = CarbonSimulatorUI(verbose=True)
    Sim.add_sgl_pos("ETH", 10, 2000, 3000, pair="ETHUSDC")
    result = Sim.add_sgl_pos("ETH", 10, 3000, 2000, pair="ETHUSDC")
    assert result["success"] is True
    assert list(Sim.state()["orders"]["p_end"].values) == [3000.0, 3000.0]


def test_inverted_range_positions_2():
    """
    Derived from `passed_tests` notebook test-13
    """
    Sim = CarbonSimulatorUI(verbose=True)
    result = Sim.add_linked_pos("BAT", 10, 12.0, 10, 0, 5, 7.5, pair="BATMAN")
    assert result["success"] is True
    assert list(Sim.state()["orders"]["p_end"].values) == [12.0, 5.0]



"""
wrapper UI class for the Carbon simulation

(c) Copyright Bprotocol foundation 2022. 
Licensed under MIT
"""
__version__ = "1.0"
__date__ = "21/Nov/2022"

import itertools
from typing import Callable, Any, Tuple, Dict, List

from tabulate import tabulate

from ..order import Order
from ..pair import CarbonPair
from ..carbon_order_ui import CarbonOrderUI
from ..routers import ExactRouterX0Y0N, AlphaRouter
from decimal import Decimal
import pandas as pd
import numpy as np


class CarbonSimulatorUI:
    """
    Providing the user interface for the Carbon simulation

    :verbose:           if True (default), the simulation logger. infos out info along the way
    :pair:              if given, Simulation methods uses this pair as default; other pairs
                        can still be used, but need to be explicitly mentioned
    :raiseonerror:      if False (default), errors are caught and returned to the result dict
    :matching_method:   if "exact" (default), the exact matching algorithm is used; otherwise the "fast" algorithm
    :decimals:          the number of decimals to which the output is rounded
    """

    def __init__(
        self,
        verbose: bool = False,
        pair: Any = None,
        raiseonerror: bool = False,
        decimals: int = 6,
        matching_method: str = "exact",
    ):
        self._pos_id = itertools.count()
        self.verbose = verbose
        self.debug = False  # setting this to True enables debug output

        if isinstance(pair, CarbonPair):
            #print("[__init__] CarbonPair provided", pair)
            self.pair = pair.pair_iso
            self._carbon_pair = pair
        else:
            self.pair = pair
            self._carbon_pair = None

        self.raiseonerror = raiseonerror
        self.numtrades = 0
        self.decimals = decimals
        self.matcher = (
            AlphaRouter(verbose=False) if matching_method == 'alpha' 
            else ExactRouterX0Y0N(verbose=False)
        )        
        self.orders = {}
        self.vault = {}
        self.trades = {}

        if self.verbose:
            print(
                f"[__init__] pair={pair}, verbose={verbose}, raiseonerror={raiseonerror}"
            )

    @property
    def default_basetoken(self):
        """
        returns the default base token of the sim (only if initialised with CarbonPair)
        """
        return self._carbon_pair.tknb if self._carbon_pair else None
    tknb = default_basetoken

    @property
    def default_quotetoken(self):
        """
        returns the default quote token of the sim (only if initialised with CarbonPair)
        """
        return self._carbon_pair.tknq if self._carbon_pair else None
    tknq = default_quotetoken

    def get_carbon_pair(self, pair: str=None, tkn: str=None) -> Tuple[CarbonPair, str]:
        """
        helper functon determining the pair from function params and class defaults

        :pair:      a string determining the pair, or CarbonPair (which is simply returned)
        :tkn:       one token that is part of the pair
        :returns:   CarbonPairStatic object, or raises
        """
        if isinstance(pair, CarbonPair):
            return pair

        if pair is None:
            if self._carbon_pair:
                # no pair is given here, we have a carbon pair in the defaults -> that's it
                return self._carbon_pair
            pair = self.pair

        if pair is None:
            raise ValueError(
                "Trading pair must be provided either in function call or in simulation defaults"
            )
        return CarbonPair.from_isopair_and_tkn(pair, tkn)

    @property
    def carbon_pair(self):
        """return the default CarbonPair associated with this object, or None"""
        try:
            cp = self.get_carbon_pair()
            return cp
        except:
            return None

    def price_convention(self, pair, tkn):
        """
        gets the price convention associated with `pair`

        :pair:      a string describing the pair (eg, "ETHUSDC")
        :tkn:       any constituent token of pair; required to split the pair into its
                    token parts; no other semantic meaning
        :returns:   a string describing the price convention (eg "USDC per ETH")
        """
        carbon_pair = self.get_carbon_pair(pair, tkn)
        return carbon_pair.price_convention

    def _add_pos(
        self,
        tkn: str,
        amt: Any,
        p_lo: Any,
        p_hi: Any,
        carbon_pair: CarbonPair,
        id1: int = None,
        id2: int = None,
    ) -> int:
        """
        PRIVATE - adds a position for sale of tkn

        :tkn:           the token that is being added to the position, eg "ETH"; it is the token being sold*
        :amt:           the amount of `tkn` that is added to the position
        :p_lo:          the lower end of the range, quoted with `tkn` as numeraire
        :p_hi:          ditto upper
        :carbon_pair:   the token pair a CarbonPair object
        :id1/2:         the order ids of the two orders that make up the position; if not given, they are generated
        :returns:       the id of the order added
        """

        # create order tracking ids
        if id1 is None:
            id1 = self._posid
        if id2 is None:
            id2 = id1

        # rearrange p_lo, p_hi if need be
        if not p_lo <= p_hi:
            pp = p_lo
            p_lo = p_hi
            p_hi = pp
            print("WARNING: swapped start and end of the range")

        # enter order
        order_params = {
            "tkn": tkn,                             # the token being sold
            "y_int": amt,                           # the capacity of the curve
            "_y": amt,                              # the initial holding is of the curve (in `tkn`)
            "p_low": p_lo,                          # the lower end of the range (`tkn` numeraire); also p_b
            "p_high": p_hi,                         # the upper end of the range (`tkn` numeraire); also p_a
            "pair_name": carbon_pair.pair_iso,      # the iso name of the pair
            "pair": carbon_pair,                    # the CarbonPair object
            "id": id1,                              # the id of the new curve generated
            "linked_to_id": id2,                    # the id to which this curve is linked (=id if single curve)
        }

        self.orders[id1] = Order(**order_params)
        return id1

    def add_sellorder(
        self, tkn: str, amt: Any, p_start: Any, p_end: Any, pair: str = None
    ) -> Dict[str, Any]:
        """
        adds a sell order for tkn

        :tkn:           the token that is being added to the position, eg "ETH"; it is the token being sold*
        :amt:           the amount of `tkn` that is added to the position
        :p_start:       the start* of the range, quoted in the currency of the pair
        :p_end:         ditto end*
        :pair:          the token pair to which the position corresponds, eg "ETHUSD"*

        *p_start, p_end are interchangeable, the code deals with sorting them correctly,
        albeit with a warning message if they were in the wrong order
        """
        try:

            # validate tkn, pair and get CarbonPair object (raises if tkn not part of pair)
            carbon_pair = self.get_carbon_pair(pair, tkn)

            # create order tracking ids
            id1 = self._posid

            # convert the prices from the pair numeraire to the curve numeraire
            p_start_c = carbon_pair.convert_price(p_start, tkn)
            p_end_c = carbon_pair.convert_price(p_end, tkn)

            # ugly hack, but somehow we need to switch lo and hi to have the
            # boundary closer to the money always first
            order_id = self._add_pos(
                tkn, amt, p_end_c, p_start_c, carbon_pair, id1, id1
            )

            if self.verbose:
                print(
                    f"[add_sgl_pos] added position: tkn={tkn}, amt={amt}, p_start={p_start}, p_end={p_end}, "
                    f"pair={carbon_pair.pair_iso}"
                )

            orders = self._to_pandas(self.orders[order_id], decimals=self.decimals)
            orderuis = {order_id: CarbonOrderUI.from_order(self.orders[order_id])}

        except Exception as e:
            if self.raiseonerror:
                raise
            return {"success": False, "error": str(e), "exception": e}
        return {"success": True, "orders": orders, "orderuis": orderuis}
    add_sgl_pos = add_sellorder
    add_order = add_sellorder

    def add_strategy(
        self,
        tkn: str,
        amt_sell: Any,
        psell_start: Any,
        psell_end: Any,
        amt_buy: Any,
        pbuy_start: Any,
        pbuy_end: Any,
        pair: str = None,
    ) -> Dict[str, Any]:
        """
        adds two linked orders (one buy, one sell; aka a "strategy")

        :tkn:           the token that is sold in the range psell_start/_end, eg "ETH"*
        :amt_sell:      the amount of `tkn` that is available for sale in range psell_start/psell_end
        :psell_start:   start of the sell `tkn` range*, quoted in the price convention of `pair`
        :psell_end:     ditto end
        :amt_buy:       the amount of the other token that is available for selling against tkn in range pbuy_start
        :pbuy_start:    start of the of the buy `tkn` range*, quoted in the price convention of `pair`
        :pbuy_end:      ditto end
        :pair:          the token pair to which the strategy corresponds, eg "ETHUSD"

        *px_start, px_end are interchangeable, the code deals with sorting them, albeit issuing a warning
        message if they are in the wrong order; sell and buy is seen from the perspective of the AMM, which
        is the same as that of the strategy (liquidity) provider
        """
        try:

            # validate tkn, pair and get CarbonPair object (raises if tkn not part of pair)
            carbon_pair = self.get_carbon_pair(pair, tkn)
            tkn2 = carbon_pair.other(tkn)

            # create order tracking ids
            id1 = self._posid
            id2 = self._posid

            # convert the prices from the pair numeraire to the curve numeraire
            psell_start_c = carbon_pair.convert_price(psell_start, tkn)
            psell_end_c = carbon_pair.convert_price(psell_end, tkn)
            pbuy_start_c = carbon_pair.convert_price(pbuy_start, tkn2)
            pbuy_end_c = carbon_pair.convert_price(pbuy_end, tkn2)

            # ugly hack but the 1/2 range boundaries are in the wrong order
            # we want the closer boundary first
            self._add_pos(
                tkn, amt_sell, psell_end_c, psell_start_c, carbon_pair, id1, id2
            )
            self._add_pos(
                tkn2, amt_buy, pbuy_end_c, pbuy_start_c, carbon_pair, id2, id1
            )

            if self.verbose:
                print(
                    f"[add_linked_pos] added tkn={tkn}, amt_sell={amt_sell}, psell_start={psell_start}, "
                    f"psell_end={psell_end}, amt_buy={amt_buy}, pbuy_start={pbuy_start}, pbuy_end={pbuy_end}, "
                    f"pair={carbon_pair.pair_iso}"
                )

            orders = pd.concat(
                [
                    self._to_pandas(o, decimals=self.decimals)
                    for o in [self.orders[id1], self.orders[id2]]
                ]
            )
            orderuis = {
                id1: CarbonOrderUI.from_order(self.orders[id1]),
                id2: CarbonOrderUI.from_order(self.orders[id2]),
            }

        except Exception as e:
            if self.raiseonerror:
                raise
            return {"success": False, "error": str(e), "exception": e}
        return {"success": True, "orders": orders, "orderuis": orderuis}
    add_linked_pos = add_strategy

    def delete_order(self, position_id):
        """
        deletes a positon / order or strategy

        :position_id:   the index of (one of) the orders to be deleted; this order, and all
                        orders linked to it will be removed
        :returns:       the orders removed as list, [] if not found
        """

        try:
            order = self.orders[position_id]
        except KeyError as e:
            if self.verbose:
                print(f"Position {position_id} does not exist: {e}")
            return []

        linked_position_id = order.linked_to_id
        if linked_position_id == position_id:
            linked_position_id = None

        pair = order.pair

        del self.orders[position_id]

        if linked_position_id:
            del self.orders[linked_position_id]

        if self.verbose:
            str2 = (
                f"and linked position {linked_position_id}"
                if linked_position_id
                else ""
            )
            print(f"Deleted position {position_id} {str2}[{pair}]")

        return (
            [position_id, linked_position_id] if linked_position_id else [position_id]
        )
    delete_pos = delete_order
    delete_strategy = delete_order

    def _trade(
        self,
        tkn: str,
        amt: Any,
        carbon_pair: CarbonPair,
        trade_action: Callable = None,
        trade_description: str = None,
        execute: bool = True,
        limit_price: Any = None,
        inpair: bool = True,
        use_positions: List[int] = None,
        threshold_orders: Any = None,
        use_positions_matchlevel: List[int] = [],

    ) -> Dict[str, Any]:
        """
        PRIVATE - executes a trade

        :tkn:                   the token that is being SOLD by the AMM, eg "ETH"*
        :amt:                   the amount to be traded*
        :carbon_pair:           the CarbonPair class of the token pair
        :trade_action:          either `match_by_target(...)` or `match_by_src(...)` method of the `self.matcher`
        :trade_description:     human-readable description of the trade, eg "amm vs trader, buy vs sell, for amt"
        :execute:               if True (default), the trade is executed; otherwise only routing is shown
        :limit_price:           the limit price of the order (this price or better from point of view
                                of the trader, not the AMM!), quoted in convention of the pair
        :inpair:                if True, only match within pair; if False (default), route through all available pairs
        :use_positions:         the positions to use for the trade (default: all positions)
        :threshold_orders:      the maximum number of order to be routed through using the alpha router

        *amt is always effectively a positive amount; however, if `trade_action` is `match_by_target` then it must
        be provided as a negative number, and if `match_by_src` as positive number
        """
        trades, orders = None, None
        try:
            decimals = self.decimals
            carbon_pair_r = carbon_pair.reverse

            if not inpair:
                raise NotImplementedError(
                    "Currently only inpair routing implemented", inpair
                )

            # parse applicable orders, indexes, and map the router ids to the simulator ids
            order_ids = [
                k
                for k in self.orders.keys()
                if self.orders[k].pair_name
                in [carbon_pair.pair_iso, carbon_pair_r.pair_iso]
                and self.orders[k].tkn == tkn
                if self.orders[k].y > 0
            ]
            use_positions = use_positions if use_positions is not None else order_ids
            applicable_orders = [v for k, v in self.orders.items() if k in order_ids and k in use_positions]
            id_map = {i: applicable_orders[i].id for i in range(len(applicable_orders))}

            if self.debug:
                print("[_trade] order_ids", order_ids)
                print("[_trade] id_map", id_map)
                print("[_trade] applicable_orders", applicable_orders)

            # set the router to only consider the applicable orders
            self.matcher.orders = applicable_orders

            # route the trade
            routes = trade_action(
                x = amt,
                threshold_orders=threshold_orders
                )

            # retrieve the final route
            final_route = routes[-1]

            #   :tkn:           the token that has been provided to the function
            #   :tkno:          the other token in the pair
            #   :tknb:          the base token of the pair
            #   :tknq:          the quote token of the pair
            tkno = carbon_pair.other(tkn)
            tknb = carbon_pair.tknb
            tknq = carbon_pair.tknq

            # :output:      the asset SOLD by the AMM, and BOUGHT by the TRADER
            #               output is a NEGATIVE number
            # :input:       the asset BOUGHT by the AMM, and SOLD by the TRADER
            #               input is a POSITIVE number
            raw_price = -final_route.total_output / final_route.total_input
            if self.debug:
                print("[_trade] final route", final_route)
                print(
                    f"[_trade] matchmethod={final_route.match_method} tkn={tkn} tkno={tkno}"
                )
                print(
                    f"[_trade] p={raw_price} num={tkn} (amm_in={final_route.total_input} "
                    f"amm_out= {final_route.total_output} mm={final_route.match_method})"
                )

            # the variable `tkn` always holds the token that goes OUT of the AMM
            # in other words: `total_output` is measured in `tkn` and `total_input` in `otkn`
            # therefore, the numeraire of raw_p = out/in is the token associated to out, `tkn`
            price_avg = carbon_pair.convert_price(raw_price, tkn)
            price_avg = round(float(price_avg), decimals)
            if self.debug:
                print("[_trade] final routes", routes)

            # we are using the function `limit_is_met` from CarbonPair; we need following inputs
            # :tkn:                 the reference token (see below)
            # :limit_price:         the limit price (execute there OR BETTER)*; this is simply the
            #                       `limit_price` parameter from the arguments
            # :buysell:             whether the person who submitted the order (here: the trader)
            #                       BUYs or SELLs (see below)
            # :currrent_price:      the price at which the transaction can be executed*; this is the
            #                        `price_avg` we've just calculated
            # *in units of the pair
            #
            # It is important to understand the following symmetry: a limit BUY order for ETH
            # at 2000 USDC per ETH is the same as a limit SELL order for USDC at 2000. We have
            # seen above that `tkn` is always the token going OUT of the AMM, therefore the
            # trader who placed the limit order BUYs it. Therefore we have
            # - `tkn`[limit_is_met] = `tkn`[function argument], and
            # - `buysell = BUY`

            if limit_price is not None:
                limitfail = not carbon_pair.limit_is_met(
                    tkn, limit_price, carbon_pair.BUY, price_avg
                )
                if limitfail:
                    execute = False
            else:
                limitfail = None

            price_at = []
            for route in routes:
                price_at += [route.input]
                route.input = round(float(route.input), decimals)
                route.output = round(float(route.output), decimals)
                route.total_price = round(abs(float(sum(price_at))), decimals)

            # iterate through the routes
            ct = 0
            trade_ids = []
            numtrades = self.numtrades
            for route in routes:

                # Get the mapped simulator index
                indx = id_map[route.index]
                if self.debug:
                    print("[_trade] final route.index", route.index)
                    print("[_trade] final indx", indx)

                amt1, amt2 = abs(route.output), abs(route.input)

                route_info = trade_description.replace("$", str(amt1)).replace(
                    "*", str(amt2)
                )

                if execute:
                    self.orders[indx].y -= Decimal(amt1)
                    if self.orders[indx].linked_to_id != indx:
                        self.orders[self.orders[indx].linked_to_id].y += Decimal(amt2)
                        if (
                            self.orders[self.orders[indx].linked_to_id].y
                            > self.orders[self.orders[indx].linked_to_id].y_int
                        ):
                            # if the linked order is overfilled, update the order y_int
                            self.orders[
                                self.orders[indx].linked_to_id
                            ].y_int = self.orders[self.orders[indx].linked_to_id].y

                # track the route info
                uid = f"{numtrades}.{ct}"
                trade_ids.append(uid)
                self.trades[uid] = {
                    "uid": [uid],  # =id.sub_id
                    "id": [numtrades],  # trade id
                    "subid": [ct],  # id of route withing trade (aggr=A)
                    "note": [f"route #{indx}"],  # human-readable note about this item
                    "aggr": [False],  # True iff this is the aggregate route
                    "exec": [
                        execute
                    ],  # True iff this trade has been executed (also False on limitfail)
                    "limitfail": [
                        limitfail
                    ],  # False if limit met, True if not, None if no limit
                    "amt1": [amt1],  # the amount of `tkn1` being sold by the AMM (>0)
                    "tkn1": [tkn],  # the token being SOLD by the AMM
                    "amt2": [amt2],  # the amount of `tkn2` being bought by the AMM (>0)
                    "tkn2": [tkno],  #  the token being BOUGHT by the AMM
                    "pair": [
                        carbon_pair.pair_iso
                    ],  # the isoname of the pair, indicating the quote token (eg ETHUSDC)
                    "routeix": [
                        indx
                    ],  # for routes: the index of the route; for aggr: list of routes
                    "nroutes": [
                        1
                    ],  # the number of routes across which the trade was executed
                    "price": [
                        price_avg
                    ],  # the price amt_/amt_, in the convention of the pair
                    "p_unit": [
                        f"{tknq} per {tknb}"
                    ],  # the units in which price is quote (eg USD per ETH)
                    "threshold_orders": [
                        threshold_orders
                    ], # the maximum number of order to be routed through using the alpha router
                }
                ct += 1
                if self.debug:
                    print("[_trade] routeinfo", route_info.replace("#", str(indx)))

            # Handle base token vs quote token variable swaps for the trade
            ttl_input = round(abs(routes[-1].total_input), decimals)
            ttl_output = round(abs(routes[-1].total_output), decimals)
            price_avg = f"{price_avg}"

            amt1, amt2 = ttl_output, ttl_input

            note = f"AMM sells {amt1:.0f}{tkn} buys {amt2:.0f}{tkno}"
            num_trades = len(routes)
            order_ids = [id_map[o.index] for o in routes]
            uid = f"{numtrades}"
            trade_ids.append(uid)
            self.trades[uid] = {
                "uid": [uid],  # see comments above
                "id": [self.numtrades],
                "subid": ["A"],
                "note": [note],
                "aggr": [True],
                "exec": [execute],
                "limitfail": [limitfail],
                "amt1": [amt1],
                "tkn1": [tkn],
                "amt2": [amt2],
                "tkn2": [tkno],
                "pair": [carbon_pair.pair_iso],
                "routeix": [str(order_ids)],
                "nroutes": [num_trades],
                "price": [price_avg],
                "p_unit": [f"{tknq} per {tknb}"],
                "threshold_orders": [threshold_orders],
            }

            # Get the trade info results
            if len(self.trades) > 0:
                trades = pd.concat([pd.DataFrame(self.trades[i]) for i in self.trades])
                trades = trades[trades["uid"].isin(trade_ids)]

            if self.verbose:
                print(note, f" ({num_trades} routes)")
            self.numtrades += 1

        except Exception as e:
            if self.raiseonerror:
                raise
            return {"success": False, "error": str(e), "exception": e}
        return {"success": True, "trades": trades}

    def amm_buys(
        self,
        tkn: str,
        amt: Any,
        pair: str = None,
        execute: bool = True,
        inpair: bool = True,
        limit_price: Any = None,
        threshold_orders: Any = None,
        use_positions: List[int] = None,
        use_positions_matchlevel: List[int] = [],
    ) -> Dict[str, Any]:
        """
        the AMM buys (and the trader sells) `amt` > 0 of `tkn`

        :tkn:               the token bought by the AMM and sold by the trader, eg "ETH"
        :amt:               the amount bought by the AMM and sold by the trader (must be positive)
        :pair:              the token pair to which the trade corresponds, eg "ETHUSD"
        :execute:           if True (default), the trade is executed; otherwise only routing is shown
        :inpair:            if True, only match within pair; if False (default), route through all available pairs
        :limit_price:       the limit price of the order (this price or better from point of view
                            of the trader, not the AMM!), quoted in convention of the pair
        :threshold_orders:  the maximum number of order to be routed through using the alpha router
        :use_positions: the positions to use for the trade (default: all positions)
        """

        try:

            if amt < 0:
                print(f"[amm_buys] negative amount {amt}; calling amm_sells")
                return self.amm_sells(
                    tkn = tkn,
                    amt = -amt,
                    pair = pair,
                    execute = execute,
                    inpair = inpair,
                    limit_price = limit_price,
                    threshold_orders = threshold_orders,
                    )

            # get the token `tkn`, the other token `tkno` and the CarbonPair object
            tkn, tkno, carbon_pair = self._get_tkn_and_validate(tkn, pair)
            self._assert_position_is_valid(tkno, carbon_pair)

            return self._trade(
                tkn=tkno,
                amt=Decimal(str(amt)),
                carbon_pair=carbon_pair,
                trade_action=self.matcher.match_by_src,
                trade_description=f"trader buys $ {tkno} sells * {tkn} | AMM sells $ {tkno} buys * {tkn} via "
                f"order #",
                execute=execute,
                limit_price=limit_price,
                inpair=inpair,
                threshold_orders=threshold_orders,
                use_positions=use_positions,
                use_positions_matchlevel=use_positions_matchlevel,
            )
        except Exception as e:
            if self.raiseonerror:
                raise
            return {"success": False, "error": str(e), "exception": e}
    trader_sells = amm_buys

    def amm_sells(
        self,
        tkn: str,
        amt: Any,
        pair: str = None,
        execute: bool = True,
        inpair: bool = True,
        limit_price: Any = None,
        threshold_orders: Any = None,
        use_positions: List[int] = None,
        use_positions_matchlevel: List[int] = [],
    ) -> Dict[str, Any]:
        """
        the AMM sells (and the trader buys) `amt` > 0 of `tkn`

        :tkn:               the token sold by the AMM and bought by the trader, eg "ETH"
        :amt:               the amount sold by the AMM and bought by the trader (must be positive)
        :pair:              the token pair to which the trade corresponds, eg "ETHUSD"
        :execute:           if True (default), the trade is executed; otherwise only routing is shown
        :inpair:            if True, only match within pair; if False (default), route through all available pairs
        :limit_price:       the limit price of the order (this price or better from point of view
                            of the trader, not the AMM!), quoted in convention of the pair
        :threshold_orders:  the maximum number of order to be routed through using the alpha router 
        :use_positions: the positions to use for the trade (default: all positions)
        """
        try:

            if amt < 0:
                print(f"[amm_sells] negative amount {amt}; calling amm_buys")
                return self.amm_buys(
                    tkn = tkn,
                    amt = -amt,
                    pair = pair,
                    execute = execute,
                    inpair = inpair,
                    limit_price = limit_price,
                    threshold_orders = threshold_orders,
                    )

            # get the token `tkn`, the other token `tkno` and the CarbonPair object
            tkn, tkno, carbon_pair = self._get_tkn_and_validate(tkn, pair)
            self._assert_position_is_valid(tkn, carbon_pair)

            return self._trade(
                tkn=tkn,
                amt=-Decimal(str(amt)),
                carbon_pair=carbon_pair,
                trade_action=self.matcher.match_by_target,
                trade_description=f"trader buys $ {tkno} sells * {tkn} | AMM sells $ {tkn} buys * {tkno} via "
                f"order #",
                execute=execute,
                limit_price=limit_price,
                inpair=inpair,
                threshold_orders=threshold_orders,
                use_positions=use_positions,
                use_positions_matchlevel=use_positions_matchlevel,

            )

        except Exception as e:
            if self.raiseonerror:
                raise
            return {"success": False, "error": str(e), "exception": e}
    trader_buys = amm_sells

    @staticmethod
    def _to_pandas(order: Order, decimals: int = 6) -> pd.DataFrame:
        """
        Exports Order values for inspection...
        """
        #print("[_to_pandas]", order.pair)

        orderui = CarbonOrderUI.from_order(order)
        #print("[_to_pandas]", orderui)
        dic = {
            "id": order.id,
            "pair": order.pair_name,
            "tkn": order.tkn,
            #"y_int": round(order.y_int, decimals),
            #"y": round(order.y, decimals),
            "y_int": float(order.y_int),
            "y": float(order.y),
            "y_unit": order.tkn,
            #"p_start": round(order.pair.convert_price(order.p_high, order.tkn), decimals),
            #"p_end": round(order.pair.convert_price(order.p_low, order.tkn), decimals),
            "p_start": float(order.pair.convert_price(order.p_high, order.tkn)),
            "p_end": float(order.pair.convert_price(order.p_low, order.tkn)),
            "p_marg": orderui.p_marg,
            "p_unit": order.pair.price_convention,
            "lid": order.linked_to_id,
        }
        return pd.DataFrame(dic, index=[order.id])

    def state(self, pair: str = None, inclhistory: bool = True) -> Dict[str, Any]:
        """
        returns a state record, describing the current state of the system

        :pair:          if None (default), the entire state of the system is provided; if it is set to a specific
                        value, only state corresponding to this specific `pair` is provided
        :inclhistory:   if True (default), also provide history, in particular trade history If False, no trades
                        are provided
        :returns:       a dict with the current state of the system where the values are pandas dataframes
        """
        try:

            if len(self.orders) > 0:
                if pair is not None:
                    if type(pair) is str:
                        pair = [pair]
                    applicable_orders = [
                        order
                        for order in self.orders.values()
                        if order.pair_name in pair
                    ]
                else:
                    applicable_orders = [order for order in self.orders.values()]
                orders = pd.concat(
                    [
                        self._to_pandas(o, decimals=self.decimals)
                        for o in applicable_orders
                    ]
                )
                orderuis = {o.id: CarbonOrderUI.from_order(o) for o in applicable_orders}
        
            else:
                orders = pd.DataFrame({})
                orderuis = dict()

            if len(self.trades) > 0:
                trades = pd.concat([pd.DataFrame(self.trades[i]) for i in self.trades])
            else:
                trades = pd.DataFrame({})

            if self.verbose and False:
                print("[state.orders]")
                print(
                    tabulate(orders, headers=list(orders.columns), tablefmt="pretty"),
                )
                print("\n\n[state.trades]")
                print(
                    tabulate(trades, headers=list(trades.columns), tablefmt="pretty"),
                )
            return {
                "orders": orders,
                "orderuis": orderuis,
                "trades": trades if inclhistory else False,
            }

        except Exception as e:
            if self.raiseonerror:
                raise
            return {"success": False, "error": str(e), "exception": e}

    ASDICT = "asdict"
    ASDF = "asdf"
    def liquidity(self, format=ASDICT):
        """
        returns the aggregate liquidity positions by pair and token, in units of the respective token(!)

        :asdf:      if True, returns the results as dataframe, otherwise as dict

        NOTE: if you pairs are provided in both directions (eg ETHUSDC and USDCETH), which is bad
        practice in any case, then those liquidity positions are not aggregated and the result
        must be corrected manually
        """
        try:
            df = pd.pivot_table(
                self.state()["orders"], 
                values=["y"], 
                index=["pair", "tkn"], 
                aggfunc=np.sum
            )
        except KeyError:
            return None 

        if format == self.ASDF:
            return df
        dct = df.to_dict(orient='dict')["y"] # yields {("ETHUSDC", ETH): ...}
        pairs = set(k[0] for k in dct)  # yields {"ETHUSDC", "ETHDAI", ...}
        result = {p:{k[1]:v for k,v in dct.items() if k[0]==p} for p in pairs} # yields {"ETHUSDC": {"ETH": ..., "USDC": ...}, ...}
        if format == self.ASDICT:
            return result
        raise ValueError(f"Unknown format {format}")

    @property
    def _posid(self) -> int:
        """
        PRIVATE - auto-incrementing position id
        """
        return next(self._pos_id)

    def _get_tkn_and_validate(self, tkn: str, pair: str) -> Tuple[str, Any, CarbonPair]:
        """
        PRIVATE - handles trade tkn, pair validations
        """
        carbon_pair = self.get_carbon_pair(pair, tkn)
        return tkn, carbon_pair.other(tkn), carbon_pair

    def _assert_tkn_is_in_positions(self, tkn: str):
        """
        PRIVATE - checks if tkn has any liquid (=non-empty) positions (raises if not)
        """
        if tkn not in {o.tkn for o in self.orders.values() if o.y > 0}:
            raise ValueError(f"token {tkn} has no non-empty liquidity positions")

    def _assert_pair_is_in_positions(self, carbon_pair: CarbonPair):
        """
        PRIVATE - checks if pair has any liquid positions (raises if not)
        """
        orders = {o.pair_name for o in self.orders.values() if o.y > 0}
        if (
            carbon_pair.pair_iso not in orders
            and carbon_pair.reverse.pair_iso not in orders
        ):
            raise ValueError(f"valid liquidity positions for {carbon_pair.pair_iso}")

    def _assert_position_is_valid(self, tkn: str, carbon_pair: CarbonPair):
        """
        PRIVATE - checks if tkn + pair exist and are valid (raises if not)
        """
        self._assert_tkn_is_in_positions(tkn)
        self._assert_pair_is_in_positions(carbon_pair)

    @property
    def numpos(self) -> int:
        """number of positions"""
        return len(self.orders)

    def __repr__(self):
        pair = self._carbon_pair if self._carbon_pair else self.pair
        return f"{self.__class__.__name__}(<{self.numpos} orders, {self.numtrades} trades>, pair='{pair}')"




"""Derived from `passed_tests` notebook test-1"""
from carbon import CarbonSimulatorUI


def test_add_sgl_pos_concentrated_on_one_point():
    """
    Derived from `passed_tests` notebook test-1
    """
    Sim = CarbonSimulatorUI(pair="USDCETH", verbose=True)
    assert Sim.add_sgl_pos("ETH", 100, 2000, 2000)["orders"].to_dict() == {
        "id": {0: 0},
        "pair": {0: "USDCETH"},
        "tkn": {0: "ETH"},
        "y_int": {0: 100.0},
        "y": {0: 100.0},
        "y_unit": {0: "ETH"},
        "p_start": {0: 2000.0},
        "p_end": {0: 2000.0},
        "p_marg": {0: 2000.0000000000002},
        "p_unit": {0: "ETH per USDC"},
        "lid": {0: 0},
    }


def test_amm_cannot_buy_eth_with_no_usdc():
    """
    Derived from `passed_tests` notebook test-1
    """
    Sim = CarbonSimulatorUI(pair="USDCETH", verbose=True)
    Sim.add_sgl_pos("ETH", 100, 2000, 2000)
    result = Sim.amm_buys("ETH", 2)
    assert result["success"] is False
    assert result["error"] == "token USDC has no non-empty liquidity positions"


def test_amm_can_sell_eth_at_curve_price():
    """
    Derived from `passed_tests` notebook test-1
    """
    Sim = CarbonSimulatorUI(pair="USDCETH", verbose=True)
    Sim.add_sgl_pos("ETH", 100, 2000, 2000)
    result = Sim.amm_sells("ETH", 2)
    assert result["success"] is True
    assert result["trades"].to_dict() == {
        "uid": {0: "0"},
        "id": {0: 0},
        "subid": {0: "A"},
        "note": {0: "AMM sells 2ETH buys 0USDC"},
        "aggr": {0: True},
        "exec": {0: True},
        "limitfail": {0: None},
        "amt1": {0: 2.0},
        "tkn1": {0: "ETH"},
        "amt2": {0: 0.001},
        "tkn2": {0: "USDC"},
        "pair": {0: "USDCETH"},
        "routeix": {0: "[0]"},
        "nroutes": {0: 1},
        "price": {0: "2000.0"},
        "p_unit": {0: "ETH per USDC"},
        'threshold_orders': {0: None},
    }


def test_can_add_multiple_single_positions():
    """
    Derived from `passed_tests` notebook test-4
    """
    Sim = CarbonSimulatorUI(pair="USDCETH", verbose=True)
    test_cases = [
        ("ETH", 0.0005, 2700, 2800),
        ("ETH", 0.0005, 2000, 2700),
        ("USDC", 1000, 2700, 2800),
        ("USDC", 1000, 2000, 2700),
    ]
    for case in test_cases:
        assert Sim.add_sgl_pos(*case)["success"] is True


def test_no_pair_provided():
    """
    Derived from `passed_tests` notebook test-6
    """
    Sim = CarbonSimulatorUI(verbose=True)
    result = Sim.add_sgl_pos("ETH", 100, 2000, 2000)
    assert result["success"] is False
    assert (
        result["error"]
        == "Trading pair must be provided either in function call or in simulation defaults"
    )


def test_geometric_mean_in_one_step():
    """
    Derived from `passed_tests` notebook test-10
    """
    Sim = CarbonSimulatorUI(verbose=True)
    Sim.add_sgl_pos("ETH", 100, 2000, 3000, "ETHUSDC")
    result = Sim.amm_sells("ETH", 100, "ETHUSDC")
    assert result["success"] is True
    assert round(Sim.state()["trades"]["price"].astype(float).mean(), 4) == 2449.4897


def test_geometric_mean_in_two_steps():
    """
    Derived from `passed_tests` notebook test-10
    """
    Sim = CarbonSimulatorUI(verbose=True)
    Sim.add_sgl_pos("ETH", 100, 2000, 3000, "ETHUSDC")
    Sim.amm_sells("ETH", 50, "ETHUSDC")
    result = Sim.amm_sells("ETH", 50, "ETHUSDC")
    assert result["success"] is True
    assert round(Sim.state()["trades"]["price"].astype(float).mean(), 4) == 2449.4897



import re

from setuptools import find_packages, setup

with open("carbon/__init__.py") as fd:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE
    ).group(1)

with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()

extras_require = {}
extras_require["complete"] = sorted(set(sum(extras_require.values(), [])))

setup(
    name="carbon-simulator",
    version=version,
    author="Bancor Network",
    author_email="mike@bancor.network",
    description="""
                    Bancor python carbon and carbon library. 
                    This python package is developed and maintained by Bancor Research. 
                    It is meant to assist in the design, testing, and validating of Bancor protocol behavior.
                """,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bancorprotocol/carbon-simulator",
    install_requires=open("requirements.txt").readlines(),
    extras_require=extras_require,
    # tests_require=open("requirements-test.txt").readlines(),
    packages=find_packages(),
    include_package_data=True,
)



"""Derived from `passed_tests` notebook test-2"""
from carbon import CarbonSimulatorUI


def test_add_linked_pos_concentrated_on_one_point():
    """
    Derived from `passed_tests` notebook test-2
    """
    Sim = CarbonSimulatorUI(pair="USDCETH", verbose=True)
    assert Sim.add_linked_pos("ETH", 100, 2000, 2000, 0, 1000, 1000)[
        "orders"
    ].to_dict() == {
        "id": {0: 0, 1: 1},
        "pair": {0: "USDCETH", 1: "USDCETH"},
        "tkn": {0: "ETH", 1: "USDC"},
        "y_int": {0: 100.0, 1: 0.0},
        "y": {0: 100.0, 1: 0.0},
        "y_unit": {0: "ETH", 1: "USDC"},
        "p_start": {0: 2000.0, 1: 1000.0},
        "p_end": {0: 2000.0, 1: 1000.0},
        "p_marg": {0: 2000.0000000000002, 1: 1000.0000000000002},
        "p_unit": {0: "ETH per USDC", 1: "ETH per USDC"},
        "lid": {0: 1, 1: 0},
    }


def test_amm_cannot_buy_eth_with_no_usdc():
    """
    Derived from `passed_tests` notebook test-2
    """
    Sim = CarbonSimulatorUI(pair="USDCETH", verbose=True)
    Sim.add_linked_pos("ETH", 100, 2000, 2000, 0, 1000, 1000)
    result = Sim.amm_buys("ETH", 10)
    assert result["success"] is False
    assert result["error"] == "token USDC has no non-empty liquidity positions"


def test_amm_can_sell_eth_at_curve_price():
    """
    Derived from `passed_tests` notebook test-1
    """
    Sim = CarbonSimulatorUI(pair="USDCETH", verbose=True)
    Sim.add_linked_pos("ETH", 100, 2000, 2000, 0, 1000, 1000)
    result = Sim.amm_sells("ETH", 10)
    assert result["success"] is True
    assert result["trades"].to_dict() == {
        "uid": {0: "0"},
        "id": {0: 0},
        "subid": {0: "A"},
        "note": {0: "AMM sells 10ETH buys 0USDC"},
        "aggr": {0: True},
        "exec": {0: True},
        "limitfail": {0: None},
        "amt1": {0: 10.0},
        "tkn1": {0: "ETH"},
        "amt2": {0: 0.005},
        "tkn2": {0: "USDC"},
        "pair": {0: "USDCETH"},
        "routeix": {0: "[0]"},
        "nroutes": {0: 1},
        "price": {0: "2000.0"},
        "p_unit": {0: "ETH per USDC"},
        'threshold_orders': {0: None},
    }


def test_amm_cannot_sell_with_insufficient_liquidity_1():
    """
    Derived from `passed_tests` notebook test-3
    """
    Sim = CarbonSimulatorUI(pair="ETHUSDC", verbose=True)
    Sim.add_linked_pos("ETH", 100, 2000, 2000, 0, 1000, 1000)
    result = Sim.amm_sells("ETH", 200)
    assert result["success"] is False
    assert (
        result["error"]
        == "Insufficient liquidity across all user positions to support this trade."
    )


def test_amm_cannot_sell_with_insufficient_liquidity_2():
    """
    Derived from `passed_tests` notebook test-3
    """
    Sim = CarbonSimulatorUI(pair="ETHUSDC", verbose=True)
    Sim.add_linked_pos("ETH", 100, 2000, 2000, 0, 1000, 1000)
    Sim.amm_sells("ETH", 10)

    # Having sold 10 ETH already, it can now not sell another 90+
    result = Sim.amm_sells("ETH", 90.0001)
    assert result["success"] is False
    assert (
        result["error"]
        == "Insufficient liquidity across all user positions to support this trade."
    )


def test_amm_sells_both_directions():
    """
    Derived from `passed_tests` notebook test-12
    """
    Sim = CarbonSimulatorUI(pair="ETHUSDC", verbose=True)
    Sim.add_linked_pos("ETH", 100, 2000, 2000, 0, 1000, 1000)

    # Sell the entire ETH position. This gives us 200,000 USDC which expands the curve
    result = Sim.amm_sells("ETH", 100)
    assert result["success"] is True
    assert result["trades"]["price"].astype(float).mean() == 2000.0

    result = Sim.amm_sells("USDC", 200000)
    assert result["success"] is True
    assert result["trades"]["price"].astype(float).mean() == 1000.0
